function [ex, ex2, entropy_x, idx_delta] = compute_truncnormal_moments(mu, precision, lower, upper)
% assume that mu, precision, lower, upper are column vectors
% each row specifies mean, precision (inverse variance), lower and upper
% limit of the corresponding truncated normal
%
% see test_compute_truncnormal_moments.m for testing 

invsigma = precision .^ 0.5;
sigma = 1 ./ invsigma;
sigma2 = 1 ./ precision; % variance

term1 = (lower - mu) .* invsigma;
term2 = (upper - mu) .* invsigma;

den = normcdf(term2) - normcdf(term1);

phi_term1 = normpdf(term1')';
phi_term2 = normpdf(term2')';
term3 = (phi_term1 - phi_term2) ./ den;

ex = mu + sigma .* term3;
% NOTE: term1*sigma + 2*mu = mu+lower
ex2 = sigma2 + mu.^2 + (sigma ./ den) .* ( compute_x_times_pdfx((mu+lower), phi_term1) - ...
    compute_x_times_pdfx((mu+upper), phi_term2) );

% for numerical stability, when den is close to 0, return a delta function at whichever of lower, upper is closest to mu
idx = find((abs(den) <= 2e-16) | isnan(den));
idx_lower_closer = abs(lower(idx) - mu(idx)) < abs(upper(idx) - mu(idx));
ex(idx) = idx_lower_closer .* lower(idx) + (1 - idx_lower_closer) .* upper(idx);
if 0 && (length(idx) >= 1)
    [ex(idx) lower(idx) upper(idx) mu(idx) idx_lower_closer sigma(idx) term3(idx) sigma(idx).*term3(idx)]
    pause
end
ex2(idx) = ex(idx).^2;

if nargout > 2
    entropy_x = 0.5*log(2*pi) + log(sigma) + log(den) ...
        + 0.5 * ( compute_x_times_pdfx(term1, phi_term1) - compute_x_times_pdfx(term2, phi_term2) ) ./ den ...
        - 0.5 * (( (phi_term1 - phi_term2) ./ den ) .^ 2);
    entropy_x(idx) = 0.5 * precision(idx) .* (mu(idx)-ex(idx)).^2; % sifting property of delta function
    if ~(sum(isnan(entropy_x))==0)
        idx
        idx2 =  find(isnan(entropy_x))   
        [sigma(idx2) precision(idx2) ex2(idx2) ex(idx2) mu(idx2) den(idx2)]
        error('entropy NaN in compute_truncnormal_moments.m')
    end 
end
idx_delta = idx;
end


function op = compute_x_times_pdfx(x, pdfx)
% to prevent nan in cases where lower/upper == inf
op = x .* pdfx;
op(isinf(abs(x))) = 0;
if ~(sum(isnan(op))==0)
    idx = find(isnan(op));
    [op(idx) x(idx) pdfx(idx)]
    error('oh oh ... bad stuff happened in compute_truncnormal_moments.m: check compute_x_times_pdfx');
end
end
