function [predZ] = ord_bin(settings, x, nQueryUrls, nLabelers, nLabels, K, ...
            v, init_id, difficulty, expt)
% EM for the Ord-Binary model
        
        
Knew = K - 1;

% not sure if this initialization is reasonable
alpha = 0.5 + 0.5 * rand(nLabelers,Knew);
beta = 0.5 + 0.5 * rand(nLabelers,Knew);
pi_vec = rand(1,K);
pi_vec = pi_vec / sum(pi_vec);

r_map = ones(K,Knew);
for k = 1:(K-1)
    r_map(k,k:Knew) = 0;
end
rtilde = r_map(x(:,3),:);

cost_function_E = -Inf*ones(settings.maxIter,1);
cost_function_M = -Inf*ones(settings.maxIter,1);
cost_function_old = -inf;

for iter = 1:settings.maxIter
    
    % E-step    
    log_lambda = ones(nQueryUrls,1) * log(pi_vec);
    tmp1 = accumarray_mat(x(:,1), rtilde.*log(alpha(x(:,2),:)+settings.EPS) + ...
        (1-rtilde).*log(1-alpha(x(:,2),:)+settings.EPS), nQueryUrls);
    tmp2 = accumarray_mat(x(:,1), (1-rtilde).*log(beta(x(:,2),:)+settings.EPS) ...
        + rtilde.*log(1-beta(x(:,2),:)+settings.EPS), nQueryUrls);
    log_lambda = log_lambda + (tmp1 * r_map' + tmp2 * (1-r_map)');
    log_lambda = log_lambda - max(log_lambda,[],2)*ones(1,K);
    lambda = exp(log_lambda) + settings.EPS;
    lambda = lambda ./ (sum(lambda,2)*ones(1,K));
    
    gamma = lambda * r_map;
    
    cost_function_E(iter) = compute_q_ordbin(x,lambda,pi_vec,alpha,beta,gamma,rtilde);
    if settings.verbose >= 2
        fprintf('iter = %4d, E-step, Q = %f\n',iter,cost_function_E(iter));
    end
    
    
    
    % M-step
    pi_vec = sum(lambda,1)/nQueryUrls; 
    
    tmp3_num = accumarray_mat(x(:,2), gamma(x(:,1),:) .* rtilde, nLabelers);
    tmp3_den = accumarray_mat(x(:,2), gamma(x(:,1),:), nLabelers) + settings.EPS;
    tmp4_num = max(0, accumarray_mat(x(:,2), (1-gamma(x(:,1),:)) .* (1-rtilde), nLabelers));
    tmp4_den = max(0, accumarray_mat(x(:,2), 1-gamma(x(:,1),:), nLabelers)) + settings.EPS;
    
    alpha = tmp3_num ./ tmp3_den;
    beta = tmp4_num ./ tmp4_den;
    
    assert(all(all(alpha>=0)) && all(all(alpha<=1)));
    assert(all(all(beta>=-settings.EPS)) && all(all(beta<=(1+settings.EPS))));
    
    cost_function_M(iter) = compute_q_ordbin(x,lambda,pi_vec,alpha,beta,gamma,rtilde);
    if settings.verbose >= 2
        fprintf('iter = %4d, M-step, Q = %f\n',iter,cost_function_M(iter));
    end
    
    % check that E- and M-steps do not decrease loglikelihood (modulo numeric errors)
    assert((cost_function_E(iter)-cost_function_old) >= -settings.numeric_error)
    assert((cost_function_M(iter)-cost_function_E(iter)) >= -settings.numeric_error)
    
    if (abs(cost_function_M(iter)-cost_function_old) < settings.thresh)
        fprintf('Breaking out since del(cost function) < %f\n',settings.thresh);
        break;
    end
    cost_function_old  = cost_function_M(iter);
    
end

predZ = lambda * v;
varZ = sum(lambda .* (ones(nQueryUrls,1)*v' - predZ*ones(1,K)).^2,2);

if settings.write
    dlmwrite([settings.op_dir expt '.' 'ord_bin' '.difficulty' difficulty '.init_id' num2str(init_id) '.predGndTruth'], ...
        [(1:nQueryUrls)' predZ varZ], 'delimiter', '\t', 'precision', 10);
    
    save([settings.op_dir expt '.' 'ord_bin' '.difficulty' difficulty '.init_id' num2str(init_id) '.mat'], ...
        'predZ','varZ','alpha','beta','lambda', 'gamma', 'cost_function_E','cost_function_M');
end
end


function op = compute_q_ordbin(x,lambda,pi_vec,alpha,beta,gamma,rtilde)
EPS = 1e-15;
op = sum(lambda,1) * log(pi_vec)';
op = op - sum(sum(lambda .* log(lambda))); % entropy term
op = op + sum(sum( gamma(x(:,1),:) .* (rtilde .* log(alpha(x(:,2),:)+EPS) ...
    + (1-rtilde) .* log(1-alpha(x(:,2),:)+EPS)) )) ...
    + sum(sum( (1-gamma(x(:,1),:)) .* ((1-rtilde) .* ...
    log(beta(x(:,2),:)+EPS) + rtilde .* log(1-beta(x(:,2),:)+EPS)) ));
assert(~isnan(op));
end