function predZ = ordmultinomial_mixture(settings, x, nQueryUrls, nLabelers, nLabels, K, ...
    v, init_id, difficulty, expt)
% variational Bayesian inference for the Ordinal-discrete-mixture model
% 
% TODO: test if the code seamlessly works with arbitrary mapped v values
% currrent default values of hyperparameters have only been tested on ratings on 1:K scale


DEBUG = 0;   % prints a lot of diagnostics --- helpful for debugging
RAND_INIT = 1;
MIN_ITER = 1;

update_tau = 1;
update_eps_n = 1;
update_phi_eta_prior = 0;
update_mu_lambda_prior = 0; %1; % if update_mu_lambda_prior == 1, ensure 
                                % that range(mu) is not very small (use
                                % settings.verbose == 2 to check this)
update_w = 1;

update_alpha_beta_prior = settings.update_alpha_beta_prior;
update_del = settings.update_del;
use_mixture = settings.use_mixture;
use_ordinal_link = settings.use_ordinal_link;
update_pi_user = settings.update_pi_user;

idx_spammers = 1:round(settings.frac_spammers * nLabelers);

fprintf('update_eps_n= %d, use_mixture= %d, use_ordinal_link= %d, update_pi_user= %d\n', ...
    update_eps_n, use_mixture, use_ordinal_link, update_pi_user);

fprintf('update_alpha_beta_prior= %d, update_phi_eta_prior= %d, update_mu_lambda_prior= %d\n', ...
    update_alpha_beta_prior, update_phi_eta_prior, update_mu_lambda_prior);

% caching some computations
numratings_n = accumarray(x(:,2),ones(nLabels,1));
numratings_m = accumarray(x(:,1),ones(nLabels,1));
idx_m_nz = find(numratings_m > 0);
idx_gt_available = find(settings.true_z ~= 0);

ex = v(x(:,3));
ex2 = ex.^2;

% ensuring that this model is applied on a 1-5 scale
% TODO: modify script to work with v (mapped ordinal values)
unique_x = sort(unique(x(:,3)));
assert(min(unique_x)==1);
assert(max(unique_x)==K);
b_thresholds_init = (v(1:end-1) + v(2:end)) / 2;
if length(v) ~= length(unique_x)
    % TODO: better boundaries for this case depending on unique_x?
    b_thresholds_init = b_thresholds_init;
end
EPS2 = 0.5; % TODO: better value for generic v's?
MINVAL = min(v) - EPS2;
MAXVAL = max(v) + EPS2;
b_lower = [MINVAL; b_thresholds_init];
b_upper = [b_thresholds_init; MAXVAL];
b_lower = b_lower(:); b_upper = b_upper(:);

pi_user = ones(nLabelers, K) / K;
if update_pi_user
    fprintf('\nWARNING: update_pi_user has been set to 1\n');
    fprintf('Model could prefer to learn different pi for each user, especially if frac_spammers high\n');
    fprintf('Code does support those updates, if you really want to update_pi_user=1,\n');
    fprintf('but I would highly recommend trying update_pi_user=0 first\n');
    error('remove me if you want to set update_pi_user=1 ... but only after you have tried update_pi_user=0 :)')
    for k = 1:K
        tmp = double(x(:,3)==k) + 0;
        pi_user(:, k) = accumarray(x(:,2), tmp);
    end
    pi_user = pi_user ./ (numratings_n * ones(1,K));
    assert(mean(abs(sum(pi_user,2)-1)) < 1e-16);
end

if use_mixture
    eps_n_ = 0.5; % proportion of times drawn from non-spam component
    eps_n = eps_n_ * ones(nLabelers, 1);
    wnm = eps_n(x(:,2)); % expected value of the binary variable
    entropy_w_tmp = - wnm.*log(wnm) - (1-wnm).*log(1-wnm);
    entropy_w = sum(entropy_w_tmp);
    loglik_y = 0;
else
    eps_n_ = 0.5; % proportion of times drawn from non-spam component
    eps_n = eps_n_ * ones(nLabelers, 1);
    wnm = ones(nLabels, 1);
    entropy_w = 0;
    loglik_y = 0;
end

idx_nk_lookup = sub2ind([nLabelers, K], x(:,2), x(:,3));

switch difficulty
    case 'off'
        C = 1;
        c_map = ones(nQueryUrls,1);
    case 'query'
        C = max(x(:,4));
        c_map  = zeros(nQueryUrls,1);
        tmp = unique([x(:,1),x(:,4)],'rows');
        c_map(tmp(:,1)) = tmp(:,2);
    case 'queryUrl'
        C = nQueryUrls;
        c_map  = (1:nQueryUrls)';
end

acc_w_n = accumarray(x(:,2), wnm);
acc_w_c = accumarray(c_map(x(:,1)), wnm);
idx_c_nz = find(acc_w_c > 0);
idx_n_nz = find(acc_w_n > 0);

%hyper parameter init
%lambda_prior = 1/var(v(x(:,3))); % not coherent to use data twice.
%mu_prior =    mean(v(x(:,3)));
lambda_prior = 0.1;
mu_prior =    mean(v);
mu = mu_prior * ones(nQueryUrls,1);
lambda = lambda_prior * ones(nQueryUrls,1);

alpha_prior = 10;
beta_prior = 5;
phi_prior = 10;
eta_prior = 5;


% parameter initializations
alpha = ones(nLabelers,1)*alpha_prior;
beta = ones(nLabelers,1)*beta_prior;
if RAND_INIT
    tau = randgamma(ones(nLabelers,1)*alpha_prior)/beta_prior;
else
    tau = ones(nLabelers,1)*alpha_prior/beta_prior;
end

Elogtau = log(tau);

phi = ones(C,1)*phi_prior;
eta = ones(C,1)*eta_prior;
if RAND_INIT
    del = randgamma(ones(C,1)*phi_prior)/eta_prior;
else
    del = ones(C,1)*phi_prior/eta_prior;
end


Elogdel = log(del);

true_ordinal_loglik = -Inf*ones(settings.maxIter,1);
cost_function_M = -Inf*ones(settings.maxIter,1);

true_corr_coef = NaN*ones(settings.maxIter,1);
true_mse = Inf*ones(settings.maxIter,1);
true_zero_one_acc = Inf*ones(settings.maxIter,1);
true_mae = Inf*ones(settings.maxIter,1);

cost_function_old = -inf;
cost_function_old_tmp = -inf;


for iter = 1:settings.maxIter
    
    % variational parameter updates
    
    % sample mu
    tt = accumarray(x(:,1), wnm .* tau(x(:,2)), [nQueryUrls 1]);
    assert(sum(isnan(tt(idx_m_nz)))==0);
    lambda(idx_m_nz) = lambda_prior + del(c_map(idx_m_nz)) .* tt(idx_m_nz);
    tt = accumarray(x(:,1), wnm .* tau(x(:,2)) .* ex);
    assert(sum(isnan(tt(idx_m_nz)))==0)
    mu(idx_m_nz) = mu_prior*lambda_prior + del(c_map(idx_m_nz)) .* tt(idx_m_nz) ;
    mu(idx_m_nz) = mu(idx_m_nz) ./ lambda(idx_m_nz);
    
    if max(abs(mu)) > 10
        warning('problem with mu update')
    end
    tmp_var_r = wnm .* (max(0, ex2 - 2*ex.*mu(x(:,1)) + mu(x(:,1)).^2) + 1./lambda(x(:,1)));
    if DEBUG && (iter > MIN_ITER)
        op = compute_q_ordmultinomial_mixture(x,alpha,beta,lambda,mu,phi,eta,...
            alpha_prior,beta_prior,lambda_prior,mu_prior,phi_prior,eta_prior,...
            tau,Elogtau,del,Elogdel,tmp_var_r,c_map,nLabelers, ...
            C,idx_m_nz, entropy_x, entropy_w, loglik_y, acc_w_n, acc_w_c);
        fprintf('iter = %4d, E-step, updating mu\t\t Q = %f\n',iter,op);
        print_q_ordmultinomial()
        try
            assert((op - cost_function_old_tmp) >= -settings.numeric_error);
        catch
            print_q_ordmultinomial()
            error('variational bound decreasing in ordmultinomial_mixture.m');
        end
        cost_function_old_tmp = op;
    end
    
    
    % update x
    if use_ordinal_link
        % update xnm; compute moments of the truncated gaussian distribution
        % precision is modulated by the wnm term
        [ex, ex2, entropy_x_terms] = ...
            compute_truncnormal_moments(mu(x(:,1)), tau(x(:,2)).*del(c_map(x(:,1))), b_lower(x(:, 3)), b_upper(x(:, 3)));
        tmp_var_r = wnm .* (max(0, ex2 - 2*ex.*mu(x(:,1)) + mu(x(:,1)).^2 + 1./lambda(x(:,1))));
        entropy_x = sum(wnm .* entropy_x_terms); % entropy weighted by wnm
    else
        tmp_var_r = wnm .* (max(0, ex2 - 2*ex.*mu(x(:,1)) + mu(x(:,1)).^2 + 1./lambda(x(:,1))));
        entropy_x = 0;
    end
    if  DEBUG && (iter > MIN_ITER)
        op = compute_q_ordmultinomial_mixture(x,alpha,beta,lambda,mu,phi,eta,...
            alpha_prior,beta_prior,lambda_prior,mu_prior,phi_prior,eta_prior,...
            tau,Elogtau,del,Elogdel,tmp_var_r,c_map,nLabelers, ...
            C,idx_m_nz, entropy_x, entropy_w, loglik_y, acc_w_n, acc_w_c);
        fprintf('iter = %4d, E-step, updating x\t\t Q = %f\n',iter,op);
        print_q_ordmultinomial()
        try
            assert((op - cost_function_old_tmp) >= -settings.numeric_error);
        catch
            print_q_ordmultinomial()
            error('variational bound decreasing in ordmultinomial_mixture.m');
        end
        cost_function_old_tmp = op;
    end
    
    % update tau
    if update_tau
        alpha = alpha_prior + acc_w_n/2;
        beta = beta_prior + 0.5 * accumarray(x(:,2), del(c_map(x(:,1))) .* tmp_var_r);
        %tau(idx_n_nz) =  alpha(idx_n_nz)  ./ beta(idx_n_nz);
        tau   =  alpha  ./ beta ;
        assert(sum(tau(idx_n_nz)<=0)==0)
        Elogtau = digamma(alpha) - log(beta);
        if DEBUG && (iter > MIN_ITER)
            op = compute_q_ordmultinomial_mixture(x,alpha,beta,lambda,mu,phi,eta,...
                alpha_prior,beta_prior,lambda_prior,mu_prior,phi_prior,eta_prior,...
                tau,Elogtau,del,Elogdel,tmp_var_r,c_map,nLabelers, ...
                C,idx_m_nz, entropy_x, entropy_w, loglik_y, acc_w_n, acc_w_c);
            fprintf('iter = %4d, E-step, updating alpha, beta, tau\t\t Q = %f\n',iter,op);
            print_q_ordmultinomial()
            assert((op - cost_function_old_tmp) >= -settings.numeric_error);
            cost_function_old_tmp = op;
        end
    end
    
    % update del
    if update_del
        phi = phi_prior + acc_w_c/2;
        eta = eta_prior + 0.5 * accumarray(c_map(x(:,1)), tmp_var_r.*tau(x(:,2)));
        del(idx_c_nz) =  phi(idx_c_nz) ./ eta(idx_c_nz);
        Elogdel = digamma(phi) - log(eta);
        
        if DEBUG && (iter > MIN_ITER)
            op = compute_q_ordmultinomial_mixture(x,alpha,beta,lambda,mu,phi,eta,...
                alpha_prior,beta_prior,lambda_prior,mu_prior,phi_prior,eta_prior,...
                tau,Elogtau,del,Elogdel,tmp_var_r,c_map,nLabelers, ...
                C,idx_m_nz, entropy_x, entropy_w, loglik_y, acc_w_n, acc_w_c);
            fprintf('iter = %4d, E-step, updating phi,eta,del\t\t Q = %f\n',iter,op);
            print_q_ordmultinomial()
            assert((op - cost_function_old_tmp) >= -settings.numeric_error);
            cost_function_old_tmp = op;
        end
        
        if DEBUG
            fprintf('iter = %4d, range(tau) = (%f, %f); range(del) = (%f, %f)\n',...
                iter,min(tau(idx_n_nz)),max(tau(idx_n_nz)),min(del(idx_c_nz)),max(del(idx_c_nz)));
        end
    end
    
    
    
    
    % update hyperparameters (type-2 ML)
    
    if update_mu_lambda_prior
        mu_prior = mean(mu(idx_m_nz));
        lambda_prior = 1/mean(max(0,(mu(idx_m_nz)-mu_prior).^2 + 1./lambda(idx_m_nz)));
    end
    if DEBUG && (iter > MIN_ITER)
        op = compute_q_ordmultinomial_mixture(x,alpha,beta,lambda,mu,phi,eta,...
            alpha_prior,beta_prior,lambda_prior,mu_prior,phi_prior,eta_prior,...
            tau,Elogtau,del,Elogdel,tmp_var_r,c_map,nLabelers, ...
            C,idx_m_nz, entropy_x, entropy_w, loglik_y, acc_w_n, acc_w_c);
        fprintf('iter = %4d, M-step, updating mu_prior, lambda_prior\t Q = %f\n',iter,op);
        print_q_ordmultinomial()
        assert((op - cost_function_old_tmp) >= -settings.numeric_error);
        cost_function_old_tmp = op;
    end
    
    if update_alpha_beta_prior && update_tau
        [alpha_prior, beta_prior] = gamma_mle_newton(mean(Elogtau(idx_n_nz)), ...
            mean(tau(idx_n_nz)), length(idx_n_nz),alpha_prior, beta_prior);
        if DEBUG && (iter > MIN_ITER)
            op = compute_q_ordmultinomial_mixture(x,alpha,beta,lambda,mu,phi,eta,...
                alpha_prior,beta_prior,lambda_prior,mu_prior,phi_prior,eta_prior,...
                tau,Elogtau,del,Elogdel,tmp_var_r,c_map,nLabelers, ...
                C,idx_m_nz, entropy_x, entropy_w, loglik_y, acc_w_n, acc_w_c);
            fprintf('iter = %4d, M-step, updating alpha_prior, beta_prior\t Q = %f\n',iter,op);
            print_q_ordmultinomial()
            assert((op - cost_function_old_tmp) >= -settings.numeric_error);
            cost_function_old_tmp = op;
        end
    end
    
    if update_phi_eta_prior && update_del
        if C > 1 % disable update when difficulty=off
            [phi_prior, eta_prior] = gamma_mle_newton(mean(Elogdel(idx_c_nz)), ...
                mean(del(idx_c_nz)), length(idx_c_nz), phi_prior, eta_prior);
            if DEBUG  && (iter > MIN_ITER)
                op = compute_q_ordmultinomial_mixture(x,alpha,beta,lambda,mu,phi,eta,...
                    alpha_prior,beta_prior,lambda_prior,mu_prior,phi_prior,eta_prior,...
                    tau,Elogtau,del,Elogdel,tmp_var_r,c_map,nLabelers, ...
                    C,idx_m_nz, entropy_x, entropy_w, loglik_y, acc_w_n, acc_w_c);
                fprintf('iter = %4d, M-step, updating phi_prior, eta_prior\t Q = %f\n',iter,op);
                print_q_ordmultinomial()
                assert((op - cost_function_old_tmp) >= -settings.numeric_error);
                cost_function_old_tmp = op;
            end
        end
    end
    
    
    if use_mixture && update_w
        tmp_var_r2 = (max(0, ex2 - 2*ex.*mu(x(:,1)) + mu(x(:,1)).^2 + 1./lambda(x(:,1))));
        termA = exp(0.5* ( log(del(c_map(x(:,1)))) + log(tau(x(:,2))) ...
            - log(2*pi) - tmp_var_r2 .* del(c_map(x(:,1))) .* tau(x(:,2)) ) );
        wnm_num = termA .* eps_n(x(:,2));
        termB = pi_user(idx_nk_lookup) ;
        wnm_den = wnm_num + (1 - eps_n(x(:,2))) .* termB;
        wnm = wnm_num ./ wnm_den;
        assert(~sum(isnan(wnm)));
        
        entropy_w_tmp = - wnm.*log(wnm) - (1-wnm).*log(1-wnm);
        entropy_w_tmp((abs(wnm) < 3e-16) | (abs(wnm-1) < 3e-16)) = 0;
        entropy_w = sum(entropy_w_tmp);
        assert(~isnan(entropy_w));
        wnm_not_1 = (wnm ~= 1);
        wnm_not_0 = (wnm ~= 0);
        
        loglik_y = sum( wnm(wnm_not_0).*log(eps_n(x(wnm_not_0,2))) ) ...
            + sum( (1-wnm(wnm_not_1)).*log(1-eps_n(x(wnm_not_1,2))) ) ...
            + sum((1-wnm) .* log(pi_user(idx_nk_lookup)));
        if isnan(loglik_y)
            [      sum( wnm.*log(eps_n(x(:,2))) ); ...
                sum( (1-wnm(wnm_not_1)).*log(1-eps_n(x(wnm_not_1,2))) ); ...
                sum((1-wnm) .* log(pi_user(idx_nk_lookup)))]'
            error('loglik_y = nan');
        end
        acc_w_n = accumarray(x(:,2), wnm);
        acc_w_c = accumarray(c_map(x(:,1)), wnm);
        idx_c_nz = find(acc_w_c > 0);
        idx_n_nz = find(acc_w_n > 0);
        tmp_var_r = wnm .* (max(0, ex2 - 2*ex.*mu(x(:,1)) + mu(x(:,1)).^2) + 1./lambda(x(:,1)));
        
    end
    
    if use_mixture && update_eps_n
        eps_n = acc_w_n ./ numratings_n;
    end
    
    
    op = compute_q_ordmultinomial_mixture(x,alpha,beta,lambda,mu,phi,eta,...
        alpha_prior,beta_prior,lambda_prior,mu_prior,phi_prior,eta_prior,...
        tau,Elogtau,del,Elogdel,tmp_var_r,c_map,nLabelers, ...
        C,idx_m_nz, entropy_x, entropy_w, loglik_y, acc_w_n, acc_w_c);
    cost_function_M(iter) = op;
    
    if settings.compute_metrics
        true_ordinal_loglik(iter) = sum(compute_ordinal_loglik(settings.true_z(idx_gt_available), ...
            mu(idx_gt_available), lambda(idx_gt_available), b_lower, b_upper));
        
        [true_mse(iter), true_mae(iter), true_corr_coef(iter), true_zero_one_acc(iter)] = ...
            compute_metrics_simple(settings.true_z(idx_gt_available), mu(idx_gt_available));
        if true_mse(iter) > 10
            error('mse just blew up .. check what is going on in ordmultinomial_mixture.m')
        end
        
        if settings.verbose >= 2
            % fprintf('iter = %4d, M-step, Q = %f, ordinal-loglik= %f\n',iter,cost_function_M(iter), true_ordinal_loglik(iter));
            fprintf('iter = %4d, M-step, Q = %.2f, ordinal-loglik= %.2f, mse= %.3f, mae= %.3f, corr= %.3f, zero-one-acc= %.3f, mean eps spammers= %.3f\n', ....
                iter,cost_function_M(iter), true_ordinal_loglik(iter), ...
                true_mse(iter), true_mae(iter), true_corr_coef(iter), ...
                true_zero_one_acc(iter), mean(eps_n(idx_spammers)));
            fprintf('iter = %4d, range(tau) = (%.2f, %.2f); range(del) = (%.2f, %.2f), range(eps) = (%.2f, %.2f) min(mu)=%.2f, max(mu) = %.2f\n',...
                iter,min(tau(idx_n_nz)),max(tau(idx_n_nz)),min(del(idx_c_nz)),max(del(idx_c_nz)), ...
                min(eps_n), max(eps_n), min(mu),max(mu));
        end
    else
        if settings.verbose >= 2
            fprintf('iter = %4d, M-step, Q = %f,\n',iter,cost_function_M(iter));
        end
    end
    
    if DEBUG
        print_q_ordmultinomial()
    end
    
    if (abs(cost_function_M(iter)-cost_function_old) < settings.thresh)
        fprintf('Breaking out since del(cost function) < %f\n',settings.thresh);
        break;
    end
    
    cost_function_old  = cost_function_M(iter);
end

if settings.compute_metrics
    fprintf('iter = %4d, M-step, Q = %.2f, ordinal-loglik= %.2f, mse= %.3f, mae= %.3f, corr= %.3f, zero-one-acc= %.3f, mean eps spammers= %.3f\n', ....
        iter,cost_function_M(iter), true_ordinal_loglik(iter), ...
        true_mse(iter), true_mae(iter), true_corr_coef(iter), ...
        true_zero_one_acc(iter), mean(eps_n(idx_spammers)));
else
    fprintf('iter = %4d, M-step, Q = %.2f, mean eps spammers= %.3f\n', ....
        iter,cost_function_M(iter), mean(eps_n(idx_spammers)));
end
fprintf('iter = %4d, range(tau) = (%.2f, %.2f); range(del) = (%.2f, %.2f), range(eps) = (%.2f, %.2f)\n',...
    iter,min(tau(idx_n_nz)),max(tau(idx_n_nz)),min(del(idx_c_nz)),max(del(idx_c_nz)), ...
    min(eps_n), max(eps_n));


fprintf('mean(eps_n) = %.3f, min(mu)=%.2f, max(mu) = %.2f\n', ...
    mean(eps_n),min(mu),max(mu));

predZ = mu;
varZ = 1./lambda;

if settings.write
    dlmwrite([settings.op_dir expt '.' settings.method_final '.difficulty' difficulty ...
        '.init_id' num2str(init_id) '.predGndTruth'],...
        [(1:nQueryUrls)' predZ varZ], 'delimiter', '\t', 'precision', 10);
    
    save([settings.op_dir expt '.' settings.method_final '.difficulty' difficulty  ...
        '.init_id' num2str(init_id) '.mat'] ...
        ,'predZ','varZ','alpha','beta','tau', 'lambda', 'mu','phi','eta','del', ...
        'alpha_prior','beta_prior','lambda_prior','mu_prior','b_lower', 'b_upper', ...
        'b_thresholds_init', 'phi_prior','eta_prior','cost_function_M' );
end
end


function op = compute_q_ordmultinomial_mixture(x,alpha,beta,lambda,mu,phi,eta,...
    alpha_prior,beta_prior,lambda_prior,mu_prior,phi_prior,eta_prior,...
    tau,Elogtau,del,Elogdel,tmp_var_r,c_map,nLabelers, ...
    C,idx_m_nz, entropy_x, entropy_w, loglik_y, acc_w_n, acc_w_c)
op =  entropy_w ...
    + entropy_x ...
    + loglik_y ...
    + 0.5 * sum( acc_w_n .* Elogtau ) - 0.5 * log(2*pi) * sum(acc_w_n) ...
    - 0.5 * sum( tmp_var_r .* del(c_map(x(:,1))) .* tau(x(:,2)) ) ...
    + C*( phi_prior*log(eta_prior)-gammaln(phi_prior) ) ...
    + sum( (acc_w_c/2+phi_prior-1).* Elogdel ) - eta_prior*sum(del)  ...
    - 0.5*sum(log(lambda(idx_m_nz)))  ...
    + nLabelers*( alpha_prior*log(beta_prior)-gammaln(alpha_prior) ) + (alpha_prior-1)*sum(Elogtau) - beta_prior*sum(tau) ...
    + sum( alpha-log(beta)+gammaln(alpha)+(1-alpha).*digamma(alpha) ) ...
    + sum( phi-log(eta)+gammaln(phi)+(1-phi).*digamma(phi) ) ...
    + length(idx_m_nz)/2*log(lambda_prior) - 0.5*lambda_prior*sum( (mu(idx_m_nz)-mu_prior).^2+1./lambda(idx_m_nz) ) ;
end
