clear all; close all; clc;

startup()

WRITE = 0;
nQueries = 200;
K = 5;
EPS = 1e-15;
b_lower = [0.5 1.5 2.5 3.5 4.5];
b_upper = [1.5 2.5 3.5 4.5 5.5];
pi = ones(K, 1) / K;

difficulty = 'off';
%difficulty = 'queryUrl'

for frac_spammer = [0 0.25 0.5]% 0.75]     % fraction of spammy annotators (i.e. eps_n = 0)
    
    tag_template = ['toydata_mixture' '_' num2str(frac_spammer)];
    for init_id = 1:1
        tag = [tag_template '_' num2str(init_id)];
        
        for numAvgRatingsPerQueryUrl = 3%1:6
            
            nQueryUrls = nQueries*10;
            nLabels = numAvgRatingsPerQueryUrl * nQueryUrls;
            
            nLabelers = round(nLabels/10);
            eps_n = ones(nLabelers,1); %0.5 + 0.5 * rand(nLabelers, 1); % set in [0.5, 1] to usual annotators
            n_spammer = round(frac_spammer*nLabelers);
            y = ones(nLabels, 1);
            idx_spammer = 1:n_spammer;
            eps_n(idx_spammer) = 0;
            v = (1:K)';
            x = zeros(nLabels,4);
            z = ceil(rand(nQueryUrls,1)*K+eps);
            x(:,1) = ceil(rand(nLabels,1)*nQueryUrls+eps);
            tmp = randperm(nLabels);
            x(tmp(1:nQueryUrls),1) = (1:nQueryUrls)'; % to ensure every qu occurs atleast once
            
            x(:,2) = ceil(rand(nLabels,1)*nLabelers+eps);
            qu2q = ceil(rand(nQueryUrls,1)*nQueries+eps);
            x(:,4) = qu2q(x(:,1)); % 4th column contains queryId
            
            % set difficulty granularity
            switch difficulty
                case 'query'
                    C = nQueries;
                    diff_x = x(:, 4);
                case 'off'
                    C = 1;
                    diff_x = ones(nLabels, 1);
                case 'queryUrl'
                    C = nQueryUrls;
                    diff_x = x(:, 1);
            end
            
            if C == 1
                del = 1;
            else
                del = randgamma(ones(C,1));
            end
            tau = 10 * randgamma(ones(nLabelers,1));
            
            
            for i = 1:nLabels
                m = x(i,1);
                n = x(i,2);
                %c = x(i,4);
                c = diff_x(i);
                y(i) = rand(1) <= eps_n(n);
                if y(i) == 0
                    x(i,3) = find(sample_hist(pi, 1));
                else
                    sigma = 1/sqrt(tau(n)*del(c));
                    x_tmp = randn(1) * sigma + z(m);
                    x_tmp = max(x_tmp, b_lower(1));
                    x_tmp = min(x_tmp, b_upper(end) - EPS);
                    idx_r = find((x_tmp >= b_lower) & (x_tmp < b_upper));
                    assert(length(idx_r) == 1);
                    x(i, 3) = idx_r;
                    
                end
            end
            
            expt = [tag '_nQueries' num2str(nQueries) '_numAvgRatingsPerQueryUrl' num2str(numAvgRatingsPerQueryUrl)];
            if WRITE
                save([expt '.mat'], 'x', 'nQueryUrls', 'nLabelers', 'nLabels', 'K', 'v', 'y', 'eps_n');
                dlmwrite([expt '.noisyRatings'], ...
                    x, 'delimiter', '\t', 'precision', 10, 'roffset', 1);
                dlmwrite([expt '.gt.discrete.quId.tsv'], ...
                    [(1:nQueryUrls)' z], 'delimiter', '\t', 'precision', 10);
                dlmwrite([expt '.gt.quId.queryId.tsv'], ...
                    [(1:nQueryUrls)' z qu2q], 'delimiter', '\t', 'precision', 10);
            end
            
            
            settings = struct();
            settings.thresh = 0.1;
            settings.maxIter = 1e5;
            settings.EPS = 1e-15;
            settings.numeric_error =  0;
            settings.write = 0;
            settings.verbose = 1;
            settings.op_dir = 'results/';
            settings.difficulty = difficulty;
            settings.frac_spammers = frac_spammer;
            method_name = 'ordmultinomial_mixture';
            update_pi_user = 0;     % recommend setting update_pi_user to 0; otherwise model might fit predictions using different pi for each user
            settings.true_z = z;
            for use_mixture = 1
                for use_ordinal_link = 1
                    for  update_alpha_beta_prior = 1
                        for update_del = 1
                            %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%');
                            
                            fprintf('\n\n\n\n');
                            fprintf('expt = %s, method_name = %s, difficulty = %s\n', ...
                                expt, method_name, difficulty);
                            fprintf('use_mixture= %d, use_ordinal_link= %d, update_pi_user= %d, update_alpha_beta_prior=%d, update_del=%d\n', ...
                                use_mixture, use_ordinal_link, update_pi_user,update_alpha_beta_prior, update_del);
                            settings.use_mixture =  use_mixture;
                            settings.use_ordinal_link =   use_ordinal_link;
                            settings.update_pi_user = update_pi_user;
                            settings.update_alpha_beta_prior = update_alpha_beta_prior;
                            settings.update_del = update_del;
                            settings.compute_metrics = 1;
                            %            settings.method_final = [method_name '_mixture_' num2str(use_mixture) ...
                            %     '_ordinal_' num2str(use_ordinal_link) '_pi_' num2str(update_pi_user) ...
                            %    '_del_' num2str(update_del) '_alphabeta_' num2str(update_alpha_beta_prior)];
                            
                            predZ = ordmultinomial_mixture(settings, x, nQueryUrls, nLabelers, nLabels, K, ...
                                v, init_id, difficulty, expt);
                            [mse, mae, corr_coef, zero_one_acc, err] = compute_metrics_simple(z, predZ);
                            fprintf('mse = %.3f, mae = %.3f, corr = %.3f, zero_one_acc = %.3f\n', mse, mae, corr_coef, zero_one_acc);
                            
                        end
                    end
                end
            end
            
            switch difficulty
                case 'off'
                    C = 1;
                    c_map = ones(nQueryUrls,1);
                    %numratings_c = nLabels;
                case 'query'
                    C = max(x(:,4));
                    c_map  = zeros(nQueryUrls,1); % make sure this does not mess up things
                    tmp = unique([x(:,1),x(:,4)],'rows');
                    c_map(tmp(:,1)) = tmp(:,2);
                    %numratings_c = accumarray(x(:,4),ones(nLabels,1));
                    sum(c_map==0)
                case 'queryUrl'
                    C = nQueryUrls;
                    c_map  = (1:nQueryUrls)';
                    %numratings_c = numratings_m;
            end
            
            wnm = y;
            ex = x(:, 3);
            
            % caching some computations
            numratings_n = accumarray(x(:,2),ones(nLabels,1));
            numratings_m = accumarray(x(:,1),ones(nLabels,1));
            idx_m_nz = find(numratings_m>0);
            
            acc_w_n = accumarray(x(:,2), wnm);
            acc_w_c = accumarray(c_map(x(:,1)), wnm);
            idx_c_nz = find(acc_w_c > 0);
            idx_n_nz = find(acc_w_n > 0);
            
            lambda_prior = 1/var(x(:,3));
            mu_prior =    mean(x(:,3));
            mu = mu_prior * ones(nQueryUrls,1);
            lambda = lambda_prior * ones(nQueryUrls,1);
            
            % sample mu
            tt = accumarray(x(:,1), wnm .* tau(x(:,2)), [nQueryUrls 1]);
            assert(sum(isnan(tt(idx_m_nz)))==0);
            lambda(idx_m_nz) = lambda_prior + del(c_map(idx_m_nz)) .* tt(idx_m_nz);
            tt = accumarray(x(:,1), wnm .* tau(x(:,2)) .* ex);
            assert(sum(isnan(tt(idx_m_nz)))==0)
            mu(idx_m_nz) = mu_prior*lambda_prior + del(c_map(idx_m_nz)) .* tt(idx_m_nz) ;
            mu(idx_m_nz) = mu(idx_m_nz) ./ lambda(idx_m_nz);
            
            err = mu - z;
            fprintf('\nfrac_spammer = %.3f, metrics using ground truth values: mse = %.3f, mae=%.3f\n', ...
                frac_spammer, mean(err .^ 2), mean(abs(err)));
            
        end
    end
end
